#include "lfs_period.h"
#include "lfs_threads.h"
#include "lfs_ptrace.h"
#include "lfs_qtrace.h"

#include <linux/aquosa/qos_debug.h>
#include <sys/time.h>
#include <asm/unistd.h>

#include <string.h>

#define __USE_ISOC99
#include <math.h>

struct tracer {
	const char *name;
	void (*get_events)(unsigned long obs_time);
	void (*call)(transf_cb_t cb, double f, double *re, double *im);
};

static struct tracer tracers[] = {
	{ "ptrace", ptrace_get_events, ptrace_call_on_events, },
	{ "qtrace", qtrace_get_events, qtrace_call_on_events, },
	{ NULL, NULL, NULL, },
};

static struct tracer *cur_tracer;

static inline unsigned long timeval_sub_us(const struct timeval *ts1, const struct timeval *ts2) {
	return (unsigned long) ((ts2->tv_sec - ts1->tv_sec) * 1000000l + (ts2->tv_usec - ts1->tv_usec));
}

/// Values in Hz
double freq_min = 10.0;
double freq_max = 400.0;
double freq_delta = 0.1;
double freq_eps = 1.0;

double *transf = NULL;
int transf_size = 0;

void set_eps(double eps) {
	freq_eps = eps;
}

static void transf_cb(double val, double f, double *re, double *im)
{
	*re += cos(2*M_PI*f*val/1000000.0);
	*im += sin(2*M_PI*f*val/1000000.0);
}

static void compute_transf() {
	int i;
	transf_size = (int)((freq_max-freq_min) / freq_delta + 1);
	transf = malloc(sizeof(transf[0]) * transf_size);
	qos_chk_exit(transf != NULL);
	for (i = 0; i < transf_size; ++i) {
		double re = 0.0;
		double im = 0.0;
		double f = freq_min + i * freq_delta;

		cur_tracer->call(transf_cb, f, &re, &im);

		transf[i] = sqrt(re*re + im*im);
		qos_log_debug("transf %g Hz = %g", f, transf[i]);
	}
}

static FILE *f_ev = NULL;

static void write_events_cb(double val, double f, double *re, double *im)
{
	fprintf(f_ev, "%lu\n", (long unsigned) val);
}

static void write_events(const char *fname) {
	double re = 0.0, im = 0.0;
	f_ev = fopen(fname, "w+");
	qos_chk_do(f_ev != NULL, exit(-1));
	cur_tracer->call(write_events_cb, 0.0, &re, &im);
	fclose(f_ev);
}

static int num_events;
static void count_events_cb(double val, double f, double *re, double *im)
{
	num_events++;
}

static int count_events() {
	double re = 0.0, im = 0.0;
	num_events = 0;
	cur_tracer->call(count_events_cb, 0.0, &re, &im);
	return num_events;
}

static unsigned long transf_period_new(double n_avg, int k_max) {
	int i = 0;
	double max_acc = 0.0;
	double max_f = 0.0;
	unsigned long period = 0;
	double avg = 0.0;
	for (i = 0; i < transf_size; ++i)
		avg += transf[i];
	avg = avg / transf_size;
	for (i = 1; i < transf_size - 1; ++i) {
		if (transf[i-1] < transf[i] && transf[i] > transf[i+1] && transf[i] > n_avg * avg) {
			/* Local maximum detected: search for multiples */
			double acc = 0.0;
			double f_i = freq_min + freq_delta * i;
			double f_j = f_i;
			int k = 0;
			do {
				double g;
				for (g = f_j - freq_eps; g <= f_j + freq_eps; g += freq_delta) {
					int j = round((g - freq_min) / freq_delta);
					if (j >= 0 && j < transf_size) {
						qos_log_debug("Accumulating transform of frequency %g (%g) onto frequency %g",
							      f_j, transf[j], f_i);
						acc += transf[j];
					}
				}
				f_j += f_i;
				++k;
			} while (f_j <= freq_max && (k_max == 0 || k < k_max));
			qos_log_debug("acc=%g (max_acc=%g)", acc, max_acc);
			if (acc > max_acc) {
				max_acc = acc;
				max_f = freq_min + freq_delta * i;
				period = (unsigned long) (1000000.0 / max_f);
				qos_log_debug("New f_max=%g, acc=%g", max_f, max_acc);
			}
		}
	}
	qos_log_debug("Returning period=%lu (freq=%g)", period, max_f);
	return period;
}

/** @return Period of first harmonic (in us) */
unsigned long transf_period(void) {
	int i;
	int max_i = 0;
	double avg = transf[0];
	for (i = 1; i < transf_size; ++i) {
		avg += transf[i];
		if (transf[i] > transf[max_i])
			max_i = i;
	}
	avg /= transf_size;
	if (avg < 0.5 * transf[max_i]) {
		double f = freq_min + freq_delta * max_i;
		unsigned long period = (unsigned long) (1000000.0 / f);
		return period;
	}
	qos_log_info("Aperiodic application detected");
	return 0;
}

void write_transf(const char *fname) {
	int i;
	FILE *file = fopen(fname, "w+");
	qos_chk_do(file != NULL, exit(-1));
	for (i = 0; i < transf_size; ++i) {
		double f = freq_min + freq_delta * i;
		fprintf(file, "%g\t%g\n", f, transf[i]);
	}
	fclose(file);
}

unsigned long infer_period(unsigned long time_us, int variation,
			   const char *events_fname, const char *transf_fname) {
	unsigned long period;
	struct timeval beg_tv, end_tv;
	unsigned long us;
	double f;

	cur_tracer->get_events(time_us);
	if (events_fname)
		write_events(events_fname);

	gettimeofday(&beg_tv, NULL);
	compute_transf();
	gettimeofday(&end_tv, NULL);
	us = (end_tv.tv_sec - beg_tv.tv_sec) * 1000000 + (end_tv.tv_usec - beg_tv.tv_usec);
	qos_log_info("Time needed to compute transform from %g to %g step %g Hz of %d events: %lu us",
		     freq_min, freq_max, freq_delta, count_events(), us);

	if (transf_fname)
		write_transf(transf_fname);

	gettimeofday(&beg_tv, NULL);

	switch (variation) {
	case 0:
		period = transf_period();
		break;
	case 1:
		period = transf_period_new(0.0, 0);
		break;
	case 2:
		// Local peaks only considered if they're 20% higher than average
		// Accumulation is made at most over 10 frequency multiples
		period = transf_period_new(1.2, 10);
		break;
	default:
		qos_chk_do(0, exit(-1));
	}
	gettimeofday(&end_tv, NULL);
	us = (end_tv.tv_sec - beg_tv.tv_sec) * 1000000 + (end_tv.tv_usec - beg_tv.tv_usec);
	if (period > 0)
		f = (double)((int)((1000000 / period)/freq_delta)) * freq_delta;
	else
		f = 0.0;
	qos_log_info("Periodic application detected, with frequency %g Hz period %lu us", f, period);
	qos_log_info("Time needed to infer period from %g to %g step %g Hz: %lu us",
		     freq_min, freq_max, freq_delta, us);

	return period;
}

int config_tracer(const char *tracer)
{
	struct tracer *tp = tracers;

	while (tp->name) {
		if (!tracer || !strcmp(tp->name, tracer)) {
			cur_tracer = tp;
			return 0;
		}
		tp++;
	}

	return 1;
}
