#include "lfs_ptrace.h"
#include "lfs_threads.h"

#include <stdlib.h>
#include <stdio.h>
#include <linux/aquosa/qos_debug.h>
#include <signal.h>

struct timeval tracing_start;
struct ptraced_child *ptraced_table[HASH_SIZE];

#define SYSCALL(name) [ __NR_ ## name ] = { # name, 1 }

struct syscall_name syscalls[__NR_syscall_max] = {
	SYSCALL(poll),
	SYSCALL(select),
	SYSCALL(nanosleep),
	SYSCALL(read),
	SYSCALL(readv),
	SYSCALL(pread64),
	SYSCALL(recvfrom),
	SYSCALL(recvmsg),
	SYSCALL(getitimer),
	SYSCALL(setitimer),
	SYSCALL(gettimeofday),
	SYSCALL(rt_sigtimedwait),
	SYSCALL(time),
	SYSCALL(timer_settime),
	SYSCALL(timer_gettime),
	SYSCALL(clock_gettime),
	SYSCALL(mq_timedreceive),
	SYSCALL(mq_timedsend),
	SYSCALL(wait4),
	SYSCALL(epoll_wait),
	SYSCALL(epoll_pwait),
	SYSCALL(epoll_wait_old),
	SYSCALL(timerfd_settime),
	SYSCALL(timerfd_gettime),
};

static int signaled;
static const int be_quiet = 1;

void ptrace_call_on_events(transf_cb_t cb, double f, double *re, double *im)
{
	struct event *e;
	double val;

	do_for_each_ptrace_event(e) {
		val = e->tv1.tv_sec * 1000000 + e->tv1.tv_usec;
		cb(val, f, re, im);
	} while_for_each_ptrace_event();
}

static inline void sub_timeval(struct timeval *tv0, struct timeval *tv1)
{
	tv0->tv_sec -= tv1->tv_sec;
	tv0->tv_usec -= tv1->tv_usec;
	if (tv0->tv_usec < 0) {
		tv0->tv_usec += 1000000;
		tv0->tv_sec--;
	}
}

void log_event(struct ptraced_child *pc, struct event *event)
{
	struct log_vec *vec, **link;

	vec = pc->cur;
	if (vec->last == LOG_VEC_ENTRIES) {
		link = &vec->next;

		xcalloc(vec, sizeof(struct log_vec));

		*link = vec;
		pc->cur = vec;
	}

	sub_timeval(&event->tv0, &tracing_start);
	sub_timeval(&event->tv1, &tracing_start);
	vec->events[vec->last++] = *event;
}

static void dump_events(struct ptraced_child *pc)
{
	struct log_vec *vec;
	struct event *e;
	int i;

	for (vec = pc->log; vec; vec = vec->next) {
		for (i = 0; i < vec->last; i++) {
			e = &vec->events[i];
			printf("%d %lu %lu %lu %s\n", pc->pid,
			       e->tv0.tv_sec, e->tv0.tv_usec,
			       (e->tv1.tv_sec - e->tv0.tv_sec) * 1000000
			       + e->tv1.tv_usec - e->tv0.tv_usec,
			       syscalls[e->orig_eax].name);
		}
	}
}

static void detach_all(void);

static void dump_all_events(void)
{
#if QOS_DEBUG_LEVEL==QOS_LEVEL_DEBUG
	struct ptraced_child *pc;
	int i;

	printf("Dumping all events...\n");
	for (i = 0; i < HASH_SIZE; i++) {
		for (pc = ptraced_table[i]; pc; pc = pc->next)
			dump_events(pc);
	}
	fflush(stdout);
#endif
}

static void dump_global_events(void)
{
	struct event *e;

	do_for_each_ptrace_event(e) {
		printf("%lf\n", time_to_double(&e->tv1));
	} while_for_each_ptrace_event();
}

static void continue_all(void) {
	int i;
	for (i = 0; i < num_threads; ++i) {
		qos_log_debug("Sending SIGCONT to %d", threads[i]);
		kill(threads[i], SIGCONT);
	}
}

static void handler(int signum)
{
	qos_log_debug("caught signal, exiting");
	continue_all();
	signaled = 1;
}

static void set_timeout(unsigned long usecs)
{
	struct itimerval itv = {
		.it_interval = { 0, 0},
		.it_value = { usecs / 1000000, usecs % 1000000 }
	};

	if (signal(SIGALRM, handler) == SIG_ERR)
		err("could not attach to signal\n");
	if (setitimer(ITIMER_REAL, &itv, NULL) == -1)
		err("could not set itimer\n");
}

struct ptraced_child *find_child(pid_t pid)
{
	struct ptraced_child *pc;

	for (pc = ptraced_table[HASH(pid)]; pc; pc = pc->next) {
		if (pc->pid == pid)
			return pc;
	}

	return NULL;
}

static int ptrace_child(pid_t pid)
{
	struct ptraced_child *pc;

	if (find_child(pid))
		return -1;

	xcalloc(pc, sizeof(struct ptraced_child));
	pc->pid = pid;

	pc->next = ptraced_table[HASH(pid)];
	xcalloc(pc->log, sizeof(struct log_vec));
	pc->cur = pc->log;
	pc->cur_event.orig_eax = -1;
	ptraced_table[HASH(pid)] = pc;

	/*
	 * NOTE: should disable signals here, to avoid leaving pid in an
	 * inconsistent state if we terminate early...
	 */

	/* Lock on target. */
	ptrace(PTRACE_ATTACH, pid, NULL, NULL);

	return 0;
}

static void wait_loop(void)
{
	struct ptraced_child *pc;
	pid_t child;
	long aux;

/* 	qos_log_debug("Waiting..."); */
	child = wait(NULL);
/* 	qos_log_debug("Got event from pid %d", child); */
	if (child == -1) {
		signaled = 1;
		return;
	}

	pc = find_child(child);
	if (!pc)
		err("bad child returned from wait()\n");

	aux = pc->cur_event.orig_eax;

	if (aux != -1) {
		if (aux >= 0 && aux < __NR_syscall_max && syscalls[aux].track) {
			gettimeofday(&pc->cur_event.tv1, NULL);
			log_event(pc, &pc->cur_event);
		}
		pc->cur_event.orig_eax = -1;
	} else {
		aux = ptrace(PTRACE_PEEKUSER, child, OFFSET, NULL);
		gettimeofday(&pc->cur_event.tv0, NULL);
		pc->cur_event.orig_eax = aux;
	}

	ptrace(PTRACE_SYSCALL, child, NULL, NULL);
}

void ptrace_get_events(unsigned long usecs) {
	int i;
	struct timeval tv0;
	typedef void (*sighandler_t)(int);
	sighandler_t old_sh;

	old_sh = signal(SIGINT, handler);
	if (old_sh == SIG_ERR)
		err("could not attach to signal\n");

	gettimeofday(&tracing_start, NULL);

	if (usecs != 0)
		set_timeout(usecs);

	for (i = 0; i < num_threads; ++i)
		ptrace_child(threads[i]);

	gettimeofday(&tv0, NULL);

	while (!signaled)
		wait_loop();

	signal(SIGINT, old_sh);

	qos_log_debug("exiting...");

	detach_all();

	if (!be_quiet) {
		dump_all_events();
		dump_global_events();
	}
}

void alarm_handler(int x) {
	qos_log_debug("1 second elapsed waiting for possible child ... giving up");
}

static void detach_all(void)
{
	int i;
	pid_t child;

	continue_all();

	signal(SIGALRM, alarm_handler);
	for (i = 0; i < num_threads; ++i) {
		// At most, wait for 1 sec for children to notify further tracing events, then give-up
		alarm(1);
		child = wait(NULL);
		alarm(0);
		if (child == -1)
			break;
		ptrace(PTRACE_DETACH, child, NULL, NULL);
	}
}
